/*
 * tvprogxml
 * Copyright (c) 2003       Erik Waling <erikw@acc.umu.se>
 * Distributed under the terms of the GNU GPL. See the COPYING file
 * for a copy of the GNU Genereal Public License.
 */

#include "main.h"

void read_xmltv_data(const char *xmltvfile) {
    
    xmlDocPtr doc;
    xmlNodePtr cur;
    
    doc = xmlParseFile(xmltvfile);

    if (!doc) {
        printf("Unable to parse document!\n"); /* FIXME: Log error */
        exit(-1);
    }

    cur = xmlDocGetRootElement(doc);
    
    if (xmlStrcmp(cur->name, (const xmlChar *) "tv")) {
        printf("Not a xmltv document, root node != tv\n");
        xmlFreeDoc(doc);
    }
    cur = cur->xmlChildrenNode;
    while (cur) {
        if ((!xmlStrcmp(cur->name, (const xmlChar *)"channel"))) {
            parse_xmltv_channels(doc, cur); 
        }
        if ((!xmlStrcmp(cur->name, (const xmlChar *)"programme"))) {
            parse_xmltv_programs(doc, cur);
        }
        cur = cur->next;
    }
    xmlFreeDoc(doc);
    fix_stop_time(chan);
}
    

void parse_xmltv_channels(xmlDocPtr doc, xmlNodePtr cur) {
    
    xmlChar *cname;
    unsigned char *tconv;
    int il, ol;
    
    xmlChar *cid = xmlGetProp(cur, (const xmlChar *) "id" );
    
    cur = cur->xmlChildrenNode;

    while (cur) {
        if ((!xmlStrcmp(cur->name, (const xmlChar *)"display-name"))) {
            cname = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
        }
        cur = cur->next;
    }
    
    if (cname) {
        tconv = (unsigned char *) malloc(strlen((char *) cname)*2);
        ol = strlen((char *) cname)*2;
        il = strlen((char *) cname);
        UTF8Toisolat1(tconv, &ol, (unsigned char *) cname, &il);
        tconv[ol]=0;
    }
    
    chan = add_channel(chan, (const char *) tconv, (const char *) cid);
    
    xmlFree(cid);
    xmlFree(cname);
    free(tconv);
    return;
}

void parse_xmltv_programs(xmlDocPtr doc, xmlNodePtr cur) {
    
    unsigned char  *tconv = NULL, *dconv = NULL;
    int            il, ol;
    xmlChar *title = NULL;
    xmlChar *desc = NULL;
    
    xmlChar *channel = xmlGetProp(cur, (const xmlChar *) "channel");
    xmlChar *start = xmlGetProp(cur, (const xmlChar *) "start");
    xmlChar *stop = xmlGetProp(cur, (const xmlChar *) "stop");
    xmlChar *showview = xmlGetProp(cur, (const xmlChar *) "showview");
        
    cur = cur->xmlChildrenNode;

    while (cur || (!title && !desc)) { /* FIXME: check logic */
        if ((!xmlStrcmp(cur->name, (const xmlChar *)"title"))) {
            title = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
        }
        if ((!xmlStrcmp(cur->name, (const xmlChar *)"desc"))) {
            desc = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
        }
        cur = cur->next;
    
    }
    if (title) {
        tconv = (unsigned char *) malloc(strlen((char *) title)*2);
        ol = strlen((char *) title)*2;
        il = strlen((char *) title);
        UTF8Toisolat1(tconv, &ol, (unsigned char *) title, &il);
        tconv[ol]=0;
    }
    if (desc) {
        dconv = (unsigned char *) malloc(strlen((char *) desc)*2);
        ol = strlen((char *) desc)*2;
        il = strlen((char *) desc);
        UTF8Toisolat1(dconv, &ol, (unsigned char *) desc, &il);
        dconv[ol]=0;
    }
    
    if (stop)
        chan = add_program(chan, (const char *) channel, (const char *) tconv, (const char *) dconv, parse_xmltv_date((const char *) start), parse_xmltv_date((const char *) stop), (const char *) showview);
    else {
        
        chan = add_program(chan, (const char *) channel, (const char *) tconv, (const char *) dconv, parse_xmltv_date((const char *) start), 0, (const char *) showview);
        /*
        printf("No stop-time found for program: %s\nFix your xmltv file with 'tv_sort --by-channel'\nExiting!\n", (char *) tconv);
        exit(EXIT_FAILURE);
        */
    }
    
    free(tconv);
    free(dconv);
    xmlFree(desc);
    xmlFree(title);
    xmlFree(channel);
    xmlFree(start);
    xmlFree(stop);
    xmlFree(showview);
    return;
}
/*
time_t parse_xmltv_time(const char *datestr) {
    
    char year[5];
    char mon[3];
    char day[3];
    char hour[3];
    char min[3];
    char sec[3];
    
    struct tm tim;
    
    memset(&year, 0, sizeof(year));
    memset(&mon, 0, sizeof(mon));
    memset(&day, 0, sizeof(day));
    memset(&hour, 0, sizeof(hour));
    memset(&min, 0, sizeof(min));
    memset(&sec, 0, sizeof(sec));
    
    memcpy(&year, datestr, 4);
    memcpy(&mon, datestr+4, 2);
    memcpy(&day, datestr+6, 2);
    memcpy(&hour, datestr+8, 2);
    memcpy(&min, datestr+10, 2);
    memcpy(&sec, datestr+12, 2);

    
    tim.tm_year = atoi(year)-1900;
    tim.tm_mon = atoi(mon)-1;
    tim.tm_mday = atoi(day);
    tim.tm_hour = atoi(hour);
    tim.tm_min = atoi(min);
    tim.tm_sec = atoi(sec);
    tim.tm_isdst = 0;
    
    return mktime(&tim);
    
}
*/

/* from tvtime's xmltv.c  */
time_t parse_xmltv_timezone(const char *tzstr) {
    
    tz_map_t date_manip_timezones[] = {
        { "IDLW",    -1200, }, { "NT",      -1100, }, { "HST",     -1000, },
        { "CAT",     -1000, }, { "AHST",    -1000, }, { "AKST",     -900, },
        { "YST",      -900, }, { "HDT",      -900, }, { "AKDT",     -800, },
        { "YDT",      -800, }, { "PST",      -800, }, { "PDT",      -700, },
        { "MST",      -700, }, { "MDT",      -600, }, { "CST",      -600, },
        { "CDT",      -500, }, { "EST",      -500, }, { "ACT",      -500, },
        { "SAT",      -400, }, { "BOT",      -400, }, { "EDT",      -400, },
        { "AST",      -400, }, { "AMT",      -400, }, { "ACST",     -400, },
        { "NFT",      -330, }, { "BRST",     -300, }, { "BRT",      -300, },
        { "AMST",     -300, }, { "ADT",      -300, }, { "ART",      -300, },
        { "NDT",      -230, }, { "AT",       -200, }, { "BRST",     -200, },
        { "FNT",      -200, }, { "WAT",      -100, }, { "FNST",     -100, },
        { "GMT",         0, }, { "UT",          0, }, { "UTC",         0, },
        { "WET",         0, }, { "CET",       100, }, { "FWT",       100, },
        { "MET",       100, }, { "MEZ",       100, }, { "MEWT",      100, },
        { "SWT",       100, }, { "BST",       100, }, { "GB",        100, },
        { "WEST",        0, }, { "CEST",      200, }, { "EET",       200, },
        { "FST",       200, }, { "MEST",      200, }, { "MESZ",      200, },
        { "METDST",    200, }, { "SAST",      200, }, { "SST",       200, },
        { "EEST",      300, }, { "BT",        300, }, { "MSK",       300, },
        { "EAT",       300, }, { "IT",        330, }, { "ZP4",       400, },
        { "MSD",       300, }, { "ZP5",       500, }, { "IST",       530, },
        { "ZP6",       600, }, { "NOVST",     600, }, { "NST",       630, },
        { "JAVT",      700, }, { "CCT",       800, }, { "AWST",      800, },
        { "WST",       800, }, { "PHT",       800, }, { "JST",       900, },
        { "ROK",       900, }, { "ACST",      930, }, { "CAST",      930, },
        { "AEST",     1000, }, { "EAST",     1000, }, { "GST",      1000, },
        { "ACDT",     1030, }, { "CADT",     1030, }, { "AEDT",     1100, },
        { "EADT",     1100, }, { "IDLE",     1200, }, { "NZST",     1200, },
        { "NZT",      1200, }, { "NZDT",     1300, } };

    const int num_timezones = sizeof( date_manip_timezones ) / sizeof( tz_map_t );
    
    if( strlen( tzstr ) == 5 && (tzstr[ 0 ] == '+' || tzstr[ 0 ] == '-') ) {
        char hour[ 3 ];
        int result;

        hour[ 0 ] = tzstr[ 1 ];
        hour[ 1 ] = tzstr[ 2 ];
        hour[ 2 ] = 0;

        result = (3600 * atoi( hour )) + (60 * atoi( tzstr + 3 ));

        return (tzstr[ 0 ] == '-') ? -result : result;
    } else {
        int i;

        for( i = 0; i < num_timezones; i++ ) {
            if( !strcasecmp( tzstr, date_manip_timezones[ i ].name ) ) {
                return 60 * ((date_manip_timezones[ i ].offset % 100) +
                             ((date_manip_timezones[ i ].offset / 100) * 60));
            }
        }
    }

    return 0;
}
/* from tvtime's xmltv.c */
time_t parse_xmltv_date(const char *date) {
    char syear[ 6 ];
    char smonth[ 3 ];
    char sday[ 3 ];
    char shour[ 3 ];
    char smin[ 3 ];
    char ssec[ 3 ];
    int len;
    struct tm tm_obj;
    time_t tz;
    const char *tzstring;
    /*
     * For some reason, mktime() accepts broken-time arguments as localtime,
     * and there is no corresponding UTC function. *Sigh*.
     * For this reason we have to calculate the offset from GMT to adjust the
     * argument given to mktime().
     */
    time_t now = time( 0 );
    long gmtoff = localtime( &now )->tm_gmtoff;

    memset( syear, 0, sizeof( syear ) );
    memset( smonth, 0, sizeof( smonth ) );
    memset( sday, 0, sizeof( sday ) );
    memset( shour, 0, sizeof( shour ) );
    memset( smin, 0, sizeof( smin ) );
    memset( ssec, 0, sizeof( ssec ) );

    /*
     * according to the xmltv dtd:
     *
     * All dates and times in this [the xmltv] DTD follow the same format,
     * loosely based on ISO 8601.  They can be 'YYYYMMDDhhmmss' or some
     * initial substring, for example if you only know the year and month you
     * can have 'YYYYMM'.  You can also append a timezone to the end; if no
     * explicit timezone is given, UT is assumed.  Examples:
     * '200007281733 BST', '200209', '19880523083000 +0300'.  (BST == +0100.)
     *
     * thus:
     * example *date = "20031022220000 +0200"
     * type:            YYYYMMDDhhmmss ZZzzz"
     * position:        0         1         2          
     *                  012345678901234567890
     *
     * note: since part of the time specification can be omitted, we cannot
     *       hard-code the offset to the timezone!
     */

    /* Find where the timezone starts */
    tzstring = strchr( date, ' ' );
    if( tzstring ) {
        /* Calculate the length of the date */
        len = tzstring - date;
        /* Parse the timezone, skipping the initial space */
        tz = parse_xmltv_timezone( tzstring + 1 );
    } else {
        /* No timezone present */
        len = strlen( date );
        /* Assume UT */
        tz = 0;
    }

    if( len >= 4 ) memcpy( syear, date, 4 );
    if( len >= 6 ) memcpy( smonth, date + 4, 2 );
    if( len >= 8 ) memcpy( sday, date + 6, 2 );
    if( len >= 10 ) memcpy( shour, date + 8, 2 );
    if( len >= 12 ) memcpy( smin, date + 10, 2 );
    if( len >= 14 ) memcpy( ssec, date + 12, 2 );

    tm_obj.tm_sec = atoi( ssec ) - tz + gmtoff;
    tm_obj.tm_min = atoi( smin );
    tm_obj.tm_hour = atoi( shour );
    tm_obj.tm_mday = atoi( sday );
    tm_obj.tm_mon = atoi( smonth ) - 1;
    tm_obj.tm_year = atoi( syear ) - 1900;
    tm_obj.tm_isdst = -1;

    return mktime( &tm_obj );
}

struct channel *add_channel(struct channel *ch, const char *channelname, const char *channelid) {

    struct channel *lp = ch;    

    if (ch) { 
        /* check if channel id already exists in lists */
        while (ch) {
            if ((strcmp(channelid, ch->channel_id)) == 0)
                return lp;

            ch = ch->next;
        }
        ch = lp;        
    }
    /* add channel */
    if (ch) {
        while (ch -> next) {
            ch = ch -> next;
        }
        ch -> next = (struct channel *) malloc(sizeof(struct channel));
        memset(ch->next, 0, sizeof(struct channel));
        ch = ch -> next;
        if (channelname)
            ch -> channel_name = strdup(channelname);
        if (channelid)
            ch -> channel_id = strdup(channelid);
        ch->placement = 0;
        return lp;
    }
    else {
        ch = (struct channel *) malloc(sizeof(struct channel));
        memset(ch, 0, sizeof(struct channel));
        if (channelname)
            ch -> channel_name = strdup(channelname);
        if (channelid)
            ch -> channel_id = strdup(channelid);
        ch->placement = 0;
        return ch;
    }

}    

struct channel *add_program(struct channel *ch, const char *channelid, const char *title, const char *desc, time_t start, time_t stop, const char *showview) {

    struct channel *lp = ch;    
    struct content *cp;
    int ret; 
    time_t t = time(NULL);
    
    /* find the right channel */
    if (ch)
        while (ch && (ret = strcmp(channelid, ch->channel_id) != 0))
             ch = ch->next;
    if (ret != 0) /* log error */
        return lp;
     
    cp = ch -> content;
   
    /* check if the program is "old" */
    if (t > stop+(conf.free_after_hours*3600))
        return lp;

    /* Check if there already is a program with the same start time.
     * In that case we won't add any new conflicting program.
     * */
    
    if (ch -> content) { 
        while (ch -> content) {
            if (ch -> content -> start_time == start) {
                ch -> content = cp;
                return lp;
            }

            ch -> content = ch -> content -> next;
        }
        ch -> content = cp;        
    }

    /* add program */
    if (ch -> content) {
        while (ch -> content -> next) {
            ch -> content = ch -> content -> next;
        }
        ch -> content -> next = (struct content *) malloc(sizeof(struct content));
        memset(ch->content->next, 0, sizeof(struct content));
        ch -> content = ch -> content -> next;
        if (title)
            ch -> content -> title = strdup(title);
        if (desc)
            ch -> content -> desc = strdup(desc);
        if (showview)
            ch -> content -> showview = strdup(showview);
        
        ch -> content -> start_time = start;
        ch -> content -> end_time = stop;
        
        ch -> content -> watch = 0;

        ch -> content -> done_alarm = 0;
        ch -> content = cp;
    }
    else {
        ch -> content = (struct content *) malloc(sizeof(struct content));
        memset(ch->content, 0, sizeof(struct content));
        if (title)
            ch -> content -> title = strdup(title);
        if (desc)
            ch -> content -> desc = strdup(desc);
        if (showview)
            ch -> content -> showview = strdup(showview);
        
        ch -> content -> start_time = start;
        ch -> content -> end_time = stop;
        
        ch -> content -> watch = 0;
        
        ch -> content -> done_alarm = 0;
    }
    return lp;
}

void debug_watch_print(struct watches *w) {

    /* used for debugging */ 
    
    while (w) {
        printf("WATCH %d: %s\n", w->type ,w->regexp);
        w = w -> next;
    }
    
}
void debug_chan_print(struct channel *ch) {

    /* used for debugging */ 
    struct content *cp;
    
    while (ch) {
        printf("CHANNEL: %s\n", ch->channel_name);
        printf("CID    : %s\n", ch->channel_id);
        cp = ch->content;
        while (ch->content) {
            printf("    TITLE: %s\n",ch->content->title);
            printf("    STRT : %d\n",(int) ch->content->start_time);
            printf("    STOP : %d\n",(int) ch->content->end_time);
            if (ch->content->watch == 1)
                printf("PROGRAM MATCHES WATCH PATTERN\n");
            ch -> content = ch->content->next;
        }
        ch->content = cp;
        
        ch = ch -> next;
    }
}

char is_watch(struct watches *w, char *str, char type) {

    regex_t rexp;

    while (w) {
        if (w->type == type) {
            regcomp(&rexp, w->regexp, REG_EXTENDED | REG_ICASE);
            if ((regexec(&rexp, str, 0, NULL, 0)) == 0) {
                return 1; /* match */
            }
            regfree(&rexp);
        }

        w = w->next;
    }

    return 0; /* no match, return false */
           
    
}

void flag_watches(struct channel *ch) {

    /* used for debugging */ 
    struct content *cp;
    
    while (ch) {
        cp = ch->content;
        while (ch->content) {
            ch->content->watch = 0;
            if (is_watch(conf.watch, ch->content->title, 1))
                ch -> content -> watch = 1;
            else if (ch->content->desc && is_watch(conf.watch, ch->content->desc, 2))
                ch -> content -> watch = 1;
            ch -> content = ch->content->next;
        }
        ch->content = cp;
        
        ch = ch -> next;
    }
    
    conf.watch = free_watch_list(conf.watch); 
}

struct watches *free_watch_list(struct watches *w) {

    struct watches *prev;
    
    while (w) {
        free(w->regexp);
        prev = w;
        w = w->next;
        memset(prev, 0, sizeof(struct watches));
        free(prev);
    }
    return NULL;
}
        

void get_programs(struct channel *c) {

    time_t t = time(NULL);
    struct content *cp;
    struct content *next, *curr; 
    int    channels=0; 
    
    
    while (c) {
        cp = c->content;
        curr = NULL;
        while (c->content) {
            /* find current program */
            if ((t >= c->content->start_time) && (t < c->content->end_time)) {
                curr = c->content;
                break;    
            }
            c->content = c->content->next;
        }
        if (curr) 
            print_current_programs(curr->title, curr->start_time, channels, curr->watch);
        else
            print_current_programs(NULL, 0, channels, 0);
        
        if (conf.show_next) {
            /* next */
            c->content = cp;
            next = NULL;
            while (c->content) {
                if (c->content->start_time > t) {
                    if (!next) {
                        next = c->content;
                    }
                    if (c->content->start_time < next->start_time) {
                        next = c->content;
                    }
                }
                c->content = c->content->next;
            }
            if (next)
                print_next_programs(next->title, next->start_time, channels, next->watch);
            else
                print_next_programs(NULL, 0, channels, 0);
        }
        channels++;
        c->content = cp; /* restore pointer to first element of list */
        c = c->next;
    }
}

void get_channels(struct channel *ch) {

    int channels=0;
    
    while (ch) {
        print_channel_names(ch->channel_name,channels);
        channels++;
        ch = ch -> next;
    }
}

struct channel *sort_channels(struct channel *ch) {

    /* uuuuuuuuuuuuuuuuuuuuuuugly routine :) */
    char num = 1;
    struct channel *new = NULL;
    struct channel *new_start = new;
    struct channel *old_start = ch;
    
    while (ch) {
        if (ch->placement == num) {
            new = add_channel(new, ch->channel_name, ch->channel_id);
            new_start = new;
            while (new->next) 
                new=new->next;
            new->channel_name = ch->channel_name;
            new->channel_id = ch->channel_id;
            new->content = ch->content;
            new->placement = -1;
            new = new_start;

            num++;
            ch = old_start; 
        }
        else
            ch = ch -> next;
    }
    ch = old_start;
    while (ch) {
        if (ch->placement == 0) {
            new = add_channel(new, ch->channel_name, ch->channel_id);
            new_start = new;
            while (new->next) 
                new=new->next;
            new->channel_name = ch->channel_name;
            new->channel_id = ch->channel_id;
            new->content = ch->content;
            new->placement = -1;
            new = new_start;
            
        }
        ch = ch -> next;
    }
    ch = old_start;    
    while (ch) {
        old_start = ch;
        ch = ch->next;
        free(old_start);
    }
        
        
    /*
    while (new) {
        printf("name: %s\n",new->channel_name);
        new = new->next;
    }
    */    
    
    return new;
  
}
void free_old_programs(struct channel *ch) {

    time_t t = time(NULL);
    struct content *prev = NULL;
    struct content *start;
    
    while (ch) {
        start = ch->content;
        while (ch->content) {
            if (t > ch->content->end_time+(conf.free_after_hours*3600)) {
                prev = ch->content;
                start = ch->content->next;
            }
            ch->content = ch->content->next;
            
            if (prev) {
                free(prev->title);
                free(prev->desc);
                free(prev->showview);
                free(prev);
                prev = NULL;
            }
        }
        ch->content = start;
        /*
        printf("Kanal: %s\nPrg: %s\n\n", ch->channel_name, ch->content->title);
        */
        ch = ch->next;
    }
}

void fix_stop_time(struct channel *ch) {

    struct content *start;
    time_t *st = NULL;
    
    while (ch) {
        start = ch->content;
        while (ch->content) {
            if (ch->content->end_time == 0) {
                st = &ch->content->end_time;
            }
            ch->content = ch->content->next;
            if (st) {
                *(st) = ch->content->start_time;
            }
            st = NULL;
        }
        ch->content = start;
        ch = ch->next;
    }
}

struct channel *find_search(char *findstr, struct channel *ch) {

    regex_t rexp;
    struct content *start;
    struct channel *hits = NULL;
    
    regcomp(&rexp, findstr, REG_EXTENDED | REG_ICASE);

    while (ch) {
        start = ch->content;
        while (ch->content) {
            if ((regexec(&rexp, ch->content->title, 0, NULL, 0)) == 0) {
                hits = add_channel(hits, ch->channel_name, ch->channel_id);
                hits = add_program(hits, (const char *) ch->channel_id, (const char *) ch->content->title, (const char *) ch->content->desc, ch->content->start_time, ch->content->end_time, (const char *) ch->content->showview);
                
            }
            ch->content = ch->content->next;
            
        }
        ch->content = start;
        ch = ch->next;
    }
            
    return hits; 
}

int count_programs(struct channel *ch) {

    int count = 0;
    struct content *start;
    
    while (ch) {
        start = ch->content;
        while (ch->content) {
            count++;
            ch->content = ch->content->next;
        }
        ch->content = start;
        ch = ch->next;
    }

    return count;
}

ITEM **build_timesorted_itemlist(struct channel *list, struct strlist *data) {
   
    int num, count;
    char *buf = (char *) malloc(BUFSIZE);
    struct strlist *data_start = data;
    struct channel *chan_start;
    struct content *cont_start;
    struct content *low, *bogus = (struct content *) malloc(sizeof(struct content));
    char *channel = NULL; 
    struct tm *tim;
    ITEM **item_list;
    
    item_list = (ITEM **) calloc( (num = count_programs(list))+1 , sizeof(ITEM **) );
    
    bogus->start_time = INT_MAX;
    
    for (count=0; count < num; count++) {
        low = bogus;
        chan_start = list;
        while (list) {
            cont_start = list->content;
            while (list->content) {
                if (low->start_time >= list->content->start_time && list->content->done_alarm != 1) {
                    low = list->content;
                    free(channel);
                    channel = strdup(list->channel_name);
                }
                list->content = list->content->next;
            }
            list->content = cont_start;
            list = list->next;
        }
        tim = localtime(&low->start_time);
        sprintf(buf,"%-*.*s %.2d:%.2d  %.30s", conf.channelname_chars, conf.channelname_chars, channel, tim->tm_hour, tim->tm_min, low->title);
        data->str = strdup(buf);
        
        data->next = (struct strlist *) malloc(sizeof(struct strlist));
        memset(data->next, 0, sizeof(struct strlist));
        data = data->next;
        
        low->done_alarm = 1;
        list = chan_start;
    }
    free(buf);
    data = data_start;
    count = 0;
    while (data) {
        if (count >= MAX_ITEMS)
            break;
        if (data->str != NULL) {
            item_list[count++] = new_item(data->str, NULL);
        }
        data = data->next;
    }
    item_list[count] = (ITEM *) NULL;
    free(bogus);
    free(channel);

    return item_list; 
}

struct strlist *free_strlist(struct strlist *list) {

    struct strlist *prev;
    
    while (list->next) {
        prev = list;
        list = list->next;
        free(prev->str);
        free(prev);
    }
    return NULL;
}

